use macros::{decode_args, spawner_param_decode_args};
use num_derive::FromPrimitive;
use num_traits::FromPrimitive;
use truth::llir::RawInstr;

use crate::Param;

#[derive(Debug, FromPrimitive, Clone, Copy)]
pub enum Opcode {
  Nop = 0,

  // control flow
  Delete = 1,
  Ret = 10,
  Call = 11,
  Jump = 12,
  JumpEq = 13,
  JumpNeq = 14,
  CallAsync = 15,
  CallAsyncById = 16,
  KillAsync = 17,
  KillAllAsync = 21,

  // stack math
  StackAlloc = 40,
  PushInt = 42,
  PopInt = 43,
  PushFloat = 44,
  PopFloat = 45,
  AddInt = 50,
  AddFloat = 51,
  SubInt = 52,
  SubFloat = 53,
  MulInt = 54,
  MulFloat = 55,
  DivInt = 56,
  DivFloat = 57,
  ModInt = 58,
  EqualInt = 59,
  EqualFloat = 60,
  NotEqualInt = 61,
  NotEqualFloat = 62,
  LessInt = 63,
  LessFloat = 64,
  LessEqualInt = 65,
  LessEqualFloat = 66,
  GreaterInt = 67,
  GreaterFloat = 68,
  GreaterEqualInt = 69,
  GreaterEqualFloat = 70,
  NotInt = 71,
  NotFloat = 72,
  LogicalOr = 73,
  LogicalAnd = 74,
  BitwiseXor = 75,
  BitwiseOr = 76,
  BitwiseAnd = 77,
  DecrementInt = 78,
  Sin = 79,
  Cos = 80,
  CirclePos = 81,
  NormalizeRad = 82,
  Wait = 83,
  NegateInt = 84,
  NegateFloat = 85,
  SquareSum = 86,
  GetAngle = 87,
  SquareRoot = 88,
  Linear = 89,
  PointRotate = 90,
  FloatTime = 91,
  Math92 = 92,
  Math93 = 93,

  EnemyCreateRel = 256,
  EnemyCreateAbs = 257,
  EnemySelectAnm = 258,
  EnemySetAnmScript = 259,
  EnemyCreateMirroredRel = 260,
  EnemyCreateMirroredAbs = 261,
  EnemySetMainAnmScript = 262,
  EnemyPlayAnmRel = 263,
  EnemyPlayAnmAbs = 264,
  EnemyCreateFillerRel = 265,
  EnemyCreateFillerAbs = 266,
  EnemyCreateFillerMirroredRel = 267,
  EnemyCreateFillerMirroredAbs = 268,
  EnemyPlaySelected = 269,
  EnemySwitchAnm = 275,
  EnemyResetAnm = 276,
  EnemyInst277 = 277,
  EnemyInst278 = 278,

  MovePosAbs = 280,
  MovePosAbsTime = 281,
  MovePosRel = 282,
  MovePosRelTime = 283,
  MoveVelAbs = 284,
  MoveVelAbsTime = 285,
  MoveVelRel = 286,
  MoveVelRelTime = 287,
  MoveCircle = 288,
  MoveCircleTime = 289,
  MoveCircleRel = 290,
  MoveCircleRelTime = 291,
  MoveRandAbs = 292,
  MoveRandRel = 293,
  MoveAdd = 294,
  MoveAddRel = 295,
  MoveEllipse = 300,
  MoveEllipseRel = 301,
  MoveBezier = 305,
  MoveReset = 307,

  SetHurtbox = 320,
  SetHitbox = 321,
  FlagSet = 322,
  FlagClear = 323,
  SetMoveBounds = 324,
  ResetMoveBounds = 325,
  ClearExtraDrops = 326,
  AddExtraDrops = 327,
  SetItemDropArea = 328,
  DropItems = 329,
  SetItemReward = 330,
  SetHealth = 331,
  SetBoss = 332,
  ResetBossTimer = 333,
  SetInterrupt = 334,
  SetInvulnerable = 335,
  PlaySound = 336,
  ShakeScreen = 337,
  StartDialogue = 338,
  WaitForDialogue = 339,
  WaitForDeath = 340,
  SetTimeoutInterrupt = 341,
  StartSpell = 342,
  EndSpell = 343,
  SetChapter = 344,
  KillAllEnemies = 345,
  ProtectPlayer = 346,
  LifeMarker = 347,
  SetByDifficultyInt = 355,
  SetByDifficultyFloat = 356,
  StartSpellDifficulty = 357,
  StartSpellDifficultyM1 = 358,
  StartSpellDifficultyM2 = 359,
  SetBossLifeCount = 360,

  SpawnerReset = 400,
  SpawnerShoot = 401,
  SpawnerSetSprite = 402,
  SpawnerSetOffset = 403,
  SpawnerSetAngle = 404,
  SpawnerSetSpeed = 405,
  SpawnerSetCount = 406,
  SpawnerSetAimMode = 407,
  SpawnerSetSounds = 408,
  SpawnerSetEffects = 409,
  SpawnerCancel = 410,
  SpawnerCopy = 411,
  LaserLineCreate = 412,
  LaserInfCreate = 413,
  LaserOffset = 414,
  LaserTarget = 415,
  LaserSpeed = 416,
  LaserWidth = 417,
  LaserAngle = 418,
  LaserRotate = 419,
  BulletCancelRadius = 420,
  BulletClearRadius = 421,
  SpawnerSpeedRank3 = 422,
  SpawnerSpeedRank5 = 423,
  SpawnerSpeedRankLerp = 424,
  SpawnerCountRank3 = 425,
  SpawnerCountRank5 = 426,
  SpawnerCountRankLerp = 427,
  LaserLineCreateNoBlend = 428,
  LaserInfCreateNoBlend = 429,
  SetFloatAngleToPlayerFromPoint = 430, // cmon can we think of a better name than that???
  LaserLineExCreate = 431,
  LaserInfExCreate = 432,
  LaserLineExCreateNoBlend = 433,
  LaserInfExCreateNoBlend = 434,
  SpawnerSpeedDifficulty = 435,
  SpawnerCountDifficulty = 436,
}

#[derive(Debug)]
pub enum PrimOp {
  Add,
  Sub,
  Mul,
  Div,
  Equal,
  NotEqual,
  Less,
  LessEqual,
  Greater,
  GreaterEqual,
  Not,
  Negate,
}

#[derive(Debug)]
pub enum Op {
  Nop,
  Delete,
  Return,
  Call(String, Vec<u8>),
  Jump {
    index: usize,
    time: i32,
  },
  JumpEqual {
    index: usize,
    time: i32,
  },
  JumpNotEqual {
    index: usize,
    time: i32,
  },
  CallAsync(String, Vec<u8>),
  CallAsyncById(String, i32, Vec<u8>),
  KillAsync(i32),
  KillAllAsync,
  StackAlloc(usize),
  PushInt(Param<i32>),
  PopInt(i32),
  PushFloat(Param<f32>),
  PopFloat(i32),
  IntOp(PrimOp),
  ModuloInt,
  FloatOp(PrimOp),
  Or,
  And,
  BitwiseXor,
  BitwiseOr,
  BitwiseAnd,
  DecrementInt(i32),
  Sin,
  Cos,
  CirclePos {
    x: Param<f32>,
    y: Param<f32>,
    angle: Param<f32>,
    radius: Param<f32>,
  },
  NormalizeRad(i32),
  Wait(Param<i32>),

  EnemyCreate {
    relative: bool,
    mirrored: bool,
    filler: bool,
    subroutine: String,
    x: Param<f32>,
    y: Param<f32>,
    health: i32,
    score: i32,
    item: i32,
  },
  EnemySelectAnm {
    source: i32,
  },
  EnemySetAnmScript {
    slot: Param<i32>,
    script: Param<i32>,
  },
  EnemySetMainAnmScript {
    slot: Param<i32>,
    script: Param<i32>,
  },
  EnemyPlayAnm {
    relative: bool,
    slot: Param<i32>,
    script: Param<i32>,
  },
  EnemyPlaySelected {
    slot: Param<i32>,
  },

  MovePos {
    relative: bool,
    x: Param<f32>,
    y: Param<f32>,
  },
  MovePosTime {
    relative: bool,
    time: Param<i32>,
    mode: Param<i32>,
    x: Param<f32>,
    y: Param<f32>,
  },

  MoveVel {
    relative: bool,
    x: Param<f32>,
    y: Param<f32>,
  },
  MoveVelTime {
    relative: bool,
    time: Param<i32>,
    mode: Param<i32>,
    x: Param<f32>,
    y: Param<f32>,
  },
  MoveRand {
    relative: bool,
    time: Param<i32>,
    mode: Param<i32>,
    speed: Param<f32>,
  },
  MoveOrbit {
    relative: bool,
    initial_angle: Param<f32>,
    speed: Param<f32>,
    radius: Param<f32>,
    radius_growth: Param<f32>,
  },
  MoveOrbitTime {
    relative: bool,
    time: Param<i32>,
    mode: Param<i32>,
    radius: Param<f32>,
    radius_growth: Param<f32>,
  },
  SetHurtbox {
    x: f32,
    y: f32,
  },
  SetHitbox {
    x: f32,
    y: f32,
  },
  SetFlag(i32),
  ClearFlag(i32),
  SetMoveBounds {
    x: f32,
    y: f32,
    width: f32,
    height: f32,
  },
  ResetMoveBounds,
  ClearExtraDrops,
  AddExtraDrops {
    drop: i32,
    item: i32,
  },
  SetItemDropArea {
    width: i32,
    height: i32,
  },
  DropItems,
  SetItemReward(i32),
  SetHealth(Param<i32>),
  SetBoss(i32),
  ResetBossTimer,
  SetInterrupt {
    slot: i32,
    hp: i32,
    time: i32,
    subroutine: String,
  },
  SetInvulnerable(Param<i32>),
  PlaySound(i32),
  ShakeScreen {
    time: Param<i32>,
    // are these name correct?
    min_intensity: Param<i32>,
    max_intensity: Param<i32>,
  },
  StartDialogue(i32),
  WaitForDialogue,
  WaitForDeath,
  SetTimeoutInterrupt {
    slot: i32,
    subroutine: String,
  },
  StartSpell {
    id: i32,
    time: i32,
    bonus: i32,
    name: String,
  },
  EndSpell,
  SetChapter(i32),
  LifeMarker {
    slot: i32,
    hp: i32,
    color: i32,
  },
  SetByDifficultyInt {
    var: i32,
    easy: i32,
    normal: i32,
    hard: i32,
    lunatic: i32,
  },
  SetByDifficultyFloat {
    var: i32,
    easy: f32,
    normal: f32,
    hard: f32,
    lunatic: f32,
  },
  StartSpellDifficulty {
    id: i32,
    time: i32,
    bonus: i32,
    name: String,
  },
  StartSpellDifficultyM1 {
    id: i32,
    time: i32,
    bonus: i32,
    name: String,
  },
  StartSpellDifficultyM2 {
    id: i32,
    time: i32,
    bonus: i32,
    name: String,
  },
  SetBossLifeCount(Param<i32>),

  SpawnerReset {
    spawner: Param<i32>,
  },
  SpawnerShoot {
    spawner: Param<i32>,
  },
  SpawnerSetSprite {
    spawner: Param<i32>,
    sprite: Param<i32>,
    color: Param<i32>,
  },
  SpawnerSetOffset {
    spawner: Param<i32>,
    x: Param<f32>,
    y: Param<f32>,
  },
  SpawnerSetAngle {
    spawner: Param<i32>,
    angle: [Param<f32>; 2],
  },
  SpawnerSetSpeed {
    spawner: Param<i32>,
    speed: [Param<f32>; 2],
  },
  SpawnerSetCount {
    spawner: Param<i32>,
    count: [Param<i32>; 2],
  },
  SpawnerSetAimMode {
    spawner: Param<i32>,
    mode: Param<i32>,
  },
  SpawnerSetSounds {
    spawner: Param<i32>,
    fire_sound: Param<i32>,
    transform_sound: Param<i32>,
  },
  SpawnerSetEffects {
    spawner: Param<i32>,
    effect_index: Param<i32>,
    is_async: Param<i32>,
    id: Param<i32>,
    int_1: Param<i32>,
    int_2: Param<i32>,
    float_1: Param<f32>,
    float_2: Param<f32>,
  },
  SpawnerCancel,
  SpawnerCopy {
    dest_spawner: Param<i32>,
    src_spawner: Param<i32>,
  },
  LaserCreate {
    opcode: Opcode,
    arguments: Vec<u8>,
  },
  LaserOffset,
  LaserTarget,
  LaserSpeed,
  LaserWidth,
  LaserAngle,
  LaserRotate,
  BulletCancelRadius(f32),
  BulletClearRadius(f32),
  SpawnerSpeedRank3 {
    spawner: Param<i32>,
    speeds: [[Param<f32>; 2]; 3],
  },
  SpawnerSpeedRank5 {
    spawner: Param<i32>,
    speeds: [[Param<f32>; 2]; 5],
  },
  SpawnerSpeedRankLerp,
  SpawnerCountRank3{
    spawner: Param<i32>,
    counts: [[Param<i32>; 2]; 3],
  },
  SpawnerCountRank5{
    spawner: Param<i32>,
    counts: [[Param<i32>; 2]; 5],
  },
  SpawnerCountRankLerp,
}

#[derive(Debug)]
pub struct Instruction {
  pub time: i32,
  pub difficulty: u8,
  pub op: Op,
}

impl Instruction {
  pub fn from_raw(inst: RawInstr, offset: u32, instruction_offsets: &Vec<(u32, usize)>) -> Self {
    let param_int = |value: i32, index: u16| {
      if inst.param_mask & (1 << index) == 0 {
        Param::Value(value)
      } else {
        Param::Variable(value)
      }
    };
    let param_float = |value: f32, index: u16| {
      if inst.param_mask & (1 << index) == 0 {
        Param::Value(value)
      } else {
        Param::Variable(value as i32)
      }
    };
    let var_int = |value: i32, index: u16| {
      assert!(inst.param_mask & (1 << index) != 0, "param must be a var");
      // assert!(value <= -9932 && value >= -10000, "value must be a var");
      value
    };
    let var_float = |value: f32, index: u16| {
      assert!(inst.param_mask & (1 << index) != 0, "param must be a var");
      // assert!(
      //   value.trunc() == value && value <= -9932. && value >= -10000.,
      //   "value must be a var"
      // );
      value as i32
    };
    let offset_to_index = |value: i32| {
      let final_offset = offset.checked_add_signed(value).expect("add overflowed");
      instruction_offsets
        [instruction_offsets.binary_search_by_key(&final_offset, |entry| entry.0).expect("failed to locate offset")]
      .1
    };

    let args = inst.args_blob;
    let opcode_raw = inst.opcode;
    let opcode = Opcode::from_u16(opcode_raw).unwrap_or_else(|| todo!("failed to convert opcode {opcode_raw}"));
    let op = match opcode {
      Opcode::Nop => Op::Nop,
      Opcode::Delete => Op::Delete,
      Opcode::Ret => Op::Return,
      Opcode::Call => {
        let (name, args) = decode_args!(args, "P(bs=4)v");
        Op::Call(name, args)
      }
      Opcode::Jump => {
        let (offset, time) = decode_args!(args, "ot");

        Op::Jump {
          index: offset_to_index(offset),
          time,
        }
      }
      Opcode::JumpEq => {
        let (offset, time) = decode_args!(args, "ot");

        Op::JumpEqual {
          index: offset_to_index(offset),
          time,
        }
      }
      Opcode::JumpNeq => {
        let (offset, time) = decode_args!(args, "ot");

        Op::JumpNotEqual {
          index: offset_to_index(offset),
          time,
        }
      }
      Opcode::CallAsync => {
        let (name, args) = decode_args!(args, "P(bs=4)v");
        Op::CallAsync(name, args)
      }
      Opcode::CallAsyncById => {
        let (name, id, args) = decode_args!(args, "P(bs=4)Sv");
        Op::CallAsyncById(name, id, args)
      }
      Opcode::KillAsync => {
        let id = decode_args!(args, "S");
        Op::KillAsync(id)
      }
      Opcode::KillAllAsync => Op::KillAllAsync,
      Opcode::StackAlloc => Op::StackAlloc(decode_args!(args, "S") as usize),
      Opcode::PushInt => {
        let value = decode_args!(args, "S");
        Op::PushInt(param_int(value, 0))
      }
      Opcode::PopInt => {
        let value = decode_args!(args, "S");
        Op::PopInt(var_int(value, 0))
      }
      Opcode::PushFloat => {
        let float = decode_args!(args, "f");
        Op::PushFloat(param_float(float, 0))
      }
      Opcode::PopFloat => {
        let var = decode_args!(args, "f");
        Op::PopFloat(var_float(var, 0))
      }
      Opcode::AddInt => Op::IntOp(PrimOp::Add),
      Opcode::AddFloat => Op::FloatOp(PrimOp::Add),
      Opcode::SubInt => Op::IntOp(PrimOp::Sub),
      Opcode::SubFloat => Op::FloatOp(PrimOp::Sub),
      Opcode::MulInt => Op::IntOp(PrimOp::Mul),
      Opcode::MulFloat => Op::FloatOp(PrimOp::Mul),
      Opcode::DivInt => Op::IntOp(PrimOp::Div),
      Opcode::DivFloat => Op::FloatOp(PrimOp::Div),
      Opcode::ModInt => Op::ModuloInt,
      Opcode::EqualInt => Op::IntOp(PrimOp::Equal),
      Opcode::EqualFloat => Op::FloatOp(PrimOp::Equal),
      Opcode::NotEqualInt => Op::IntOp(PrimOp::NotEqual),
      Opcode::NotEqualFloat => Op::FloatOp(PrimOp::NotEqual),
      Opcode::LessInt => Op::IntOp(PrimOp::Less),
      Opcode::LessFloat => Op::FloatOp(PrimOp::Less),
      Opcode::LessEqualInt => Op::IntOp(PrimOp::LessEqual),
      Opcode::LessEqualFloat => Op::FloatOp(PrimOp::LessEqual),
      Opcode::GreaterInt => Op::IntOp(PrimOp::Greater),
      Opcode::GreaterFloat => Op::FloatOp(PrimOp::Greater),
      Opcode::GreaterEqualInt => Op::IntOp(PrimOp::GreaterEqual),
      Opcode::GreaterEqualFloat => Op::FloatOp(PrimOp::GreaterEqual),
      Opcode::NotInt => Op::IntOp(PrimOp::Not),
      Opcode::NotFloat => Op::FloatOp(PrimOp::Not),
      Opcode::LogicalOr => Op::Or,
      Opcode::LogicalAnd => Op::And,
      Opcode::BitwiseXor => Op::BitwiseXor,
      Opcode::BitwiseOr => Op::BitwiseOr,
      Opcode::BitwiseAnd => Op::BitwiseAnd,
      Opcode::DecrementInt => {
        let by = decode_args!(args, "S");
        Op::DecrementInt(var_int(by, 0))
      }
      Opcode::Sin => Op::Sin,
      Opcode::Cos => Op::Cos,
      Opcode::CirclePos => {
        let (x, y, angle, radius) = decode_args!(args, "ffff");
        let (x, y, angle, radius) = (
          param_float(x, 0),
          param_float(y, 1),
          param_float(angle, 2),
          param_float(radius, 3),
        );
        Op::CirclePos { x, y, angle, radius }
      }
      Opcode::NormalizeRad => {
        let var = decode_args!(args, "S");
        Op::NormalizeRad(var_int(var, 0))
      }
      Opcode::Wait => {
        let time = decode_args!(args, "S");
        Op::Wait(param_int(time, 0))
      }
      Opcode::NegateInt => Op::IntOp(PrimOp::Negate),
      Opcode::NegateFloat => Op::FloatOp(PrimOp::Negate),
      Opcode::EnemyCreateRel
      | Opcode::EnemyCreateAbs
      | Opcode::EnemyCreateMirroredRel
      | Opcode::EnemyCreateMirroredAbs
      | Opcode::EnemyCreateFillerRel
      | Opcode::EnemyCreateFillerAbs
      | Opcode::EnemyCreateFillerMirroredRel
      | Opcode::EnemyCreateFillerMirroredAbs => {
        let (subroutine, x, y, health, score, item) = decode_args!(args, "P(bs=4)ffSSS");
        let (x, y) = (param_float(x, 0), param_float(y, 1));
        Op::EnemyCreate {
          relative: matches!(
            opcode,
            Opcode::EnemyCreateRel
              | Opcode::EnemyCreateMirroredRel
              | Opcode::EnemyCreateFillerRel
              | Opcode::EnemyCreateFillerMirroredRel
          ),
          mirrored: matches!(
            opcode,
            Opcode::EnemyCreateMirroredRel
              | Opcode::EnemyCreateMirroredAbs
              | Opcode::EnemyCreateFillerMirroredRel
              | Opcode::EnemyCreateFillerMirroredAbs
          ),
          filler: matches!(
            opcode,
            Opcode::EnemyCreateFillerRel
              | Opcode::EnemyCreateFillerAbs
              | Opcode::EnemyCreateFillerMirroredRel
              | Opcode::EnemyCreateFillerMirroredAbs
          ),
          subroutine,
          x,
          y,
          health,
          score,
          item,
        }
      }
      Opcode::EnemySelectAnm => {
        let source = decode_args!(args, "S");
        Op::EnemySelectAnm { source }
      }
      Opcode::EnemySetAnmScript => {
        let (slot, script) = decode_args!(args, "SS");
        let (slot, script) = (param_int(slot, 0), param_int(script, 1));
        Op::EnemySetAnmScript { slot, script }
      }
      Opcode::EnemySetMainAnmScript => {
        let (slot, script) = decode_args!(args, "SS");
        let (slot, script) = (param_int(slot, 0), param_int(script, 1));
        Op::EnemySetMainAnmScript { slot, script }
      }
      Opcode::EnemyPlayAnmRel | Opcode::EnemyPlayAnmAbs => {
        let (slot, script) = decode_args!(args, "SS");
        let (slot, script) = (param_int(slot, 0), param_int(script, 1));
        Op::EnemyPlayAnm {
          relative: matches!(opcode, Opcode::EnemyPlayAnmRel),
          slot,
          script,
        }
      }
      Opcode::EnemyPlaySelected => {
        let slot = decode_args!(args, "S");
        Op::EnemyPlaySelected {
          slot: param_int(slot, 0),
        }
      }
      Opcode::MovePosAbs | Opcode::MovePosRel => {
        let (x, y) = decode_args!(args, "ff");
        let (x, y) = (param_float(x, 0), param_float(y, 1));
        Op::MovePos {
          relative: matches!(opcode, Opcode::MovePosRel),
          x,
          y,
        }
      }
      Opcode::MovePosAbsTime | Opcode::MovePosRelTime => {
        let (time, mode, x, y) = decode_args!(args, "SSff");
        let (time, mode, x, y) = (
          param_int(time, 0),
          param_int(mode, 1),
          param_float(x, 2),
          param_float(y, 3),
        );
        Op::MovePosTime {
          relative: matches!(opcode, Opcode::MovePosRelTime),
          time,
          mode,
          x,
          y,
        }
      }
      Opcode::MoveVelAbs | Opcode::MoveVelRel => {
        let (x, y) = decode_args!(args, "ff");
        let (x, y) = (param_float(x, 0), param_float(y, 1));
        Op::MoveVel {
          relative: matches!(opcode, Opcode::MoveVelRel),
          x,
          y,
        }
      }
      Opcode::MoveVelAbsTime | Opcode::MoveVelRelTime => {
        let (time, mode, x, y) = decode_args!(args, "SSff");
        let (time, mode, x, y) = (
          param_int(time, 0),
          param_int(mode, 1),
          param_float(x, 2),
          param_float(y, 3),
        );
        Op::MoveVelTime {
          relative: matches!(opcode, Opcode::MoveVelRelTime),
          time,
          mode,
          x,
          y,
        }
      }
      Opcode::MoveRandAbs | Opcode::MoveRandRel => {
        let (time, mode, speed) = decode_args!(args, "SSf");
        let (time, mode, speed) = (param_int(time, 0), param_int(mode, 1), param_float(speed, 2));
        Op::MoveRand {
          relative: matches!(opcode, Opcode::MoveRandRel),
          time,
          mode,
          speed,
        }
      }
      Opcode::SetHurtbox => {
        let (x, y) = decode_args!(args, "ff");
        Op::SetHurtbox { x, y }
      }
      Opcode::SetHitbox => {
        let (x, y) = decode_args!(args, "ff");
        Op::SetHitbox { x, y }
      }
      Opcode::FlagSet => {
        let flag = decode_args!(args, "S");
        Op::SetFlag(flag)
      }
      Opcode::FlagClear => {
        let flag = decode_args!(args, "S");
        Op::ClearFlag(flag)
      }
      Opcode::SetMoveBounds => {
        let (x, y, width, height) = decode_args!(args, "ffff");
        Op::SetMoveBounds { x, y, width, height }
      }
      Opcode::ResetMoveBounds => Op::ResetMoveBounds,
      Opcode::ClearExtraDrops => Op::ClearExtraDrops,
      Opcode::AddExtraDrops => {
        let (drop, item) = decode_args!(args, "SS");
        Op::AddExtraDrops { drop, item }
      }
      Opcode::SetItemDropArea => {
        let (width, height) = decode_args!(args, "SS");
        Op::SetItemDropArea { width, height }
      }
      Opcode::DropItems => Op::DropItems,
      Opcode::SetItemReward => {
        let reward = decode_args!(args, "S");
        Op::SetItemReward(reward)
      }
      Opcode::SetHealth => {
        let health = decode_args!(args, "S");
        Op::SetHealth(param_int(health, 0))
      }
      Opcode::SetBoss => {
        let boss = decode_args!(args, "S");
        Op::SetBoss(boss)
      }
      Opcode::ResetBossTimer => Op::ResetBossTimer,
      Opcode::SetInterrupt => {
        let (slot, hp, time, subroutine) = decode_args!(args, "SSSP(bs=4)");
        Op::SetInterrupt {
          slot,
          hp,
          time,
          subroutine,
        }
      }
      Opcode::SetInvulnerable => {
        let timer = decode_args!(args, "S");
        Op::SetInvulnerable(param_int(timer, 0))
      }
      Opcode::PlaySound => {
        let sound = decode_args!(args, "S");
        Op::PlaySound(sound)
      }
      Opcode::ShakeScreen => {
        let (time, min_intensity, max_intensity) = decode_args!(args, "SSS");
        let (time, min_intensity, max_intensity) = (
          param_int(time, 0),
          param_int(min_intensity, 1),
          param_int(max_intensity, 2),
        );
        Op::ShakeScreen {
          time,
          min_intensity,
          max_intensity,
        }
      }
      Opcode::StartDialogue => {
        let id = decode_args!(args, "S");
        Op::StartDialogue(id)
      }
      Opcode::WaitForDialogue => Op::WaitForDialogue,
      Opcode::WaitForDeath => Op::WaitForDeath,
      Opcode::SetTimeoutInterrupt => {
        let (slot, subroutine) = decode_args!(args, "SP(bs=4)");

        Op::SetTimeoutInterrupt { slot, subroutine }
      }
      Opcode::StartSpell => {
        let (id, time, bonus, name) = decode_args!(args, "SSSp(bs=4;mask=0x77,7,16)");
        Op::StartSpell { id, time, bonus, name }
      }
      Opcode::EndSpell => Op::EndSpell,
      Opcode::SetChapter => {
        let chapter = decode_args!(args, "S");
        Op::SetChapter(chapter)
      }
      Opcode::LifeMarker => {
        let (slot, hp, color) = decode_args!(args, "SSS");
        Op::LifeMarker { slot, hp, color }
      }
      Opcode::SetByDifficultyInt => {
        let (var, easy, normal, hard, lunatic) = decode_args!(args, "SSSSS");
        Op::SetByDifficultyInt {
          var: var_int(var, 0),
          easy,
          normal,
          hard,
          lunatic,
        }
      }
      Opcode::SetByDifficultyFloat => {
        let (var, easy, normal, hard, lunatic) = decode_args!(args, "fffff");
        Op::SetByDifficultyFloat {
          var: var_float(var, 0),
          easy,
          normal,
          hard,
          lunatic,
        }
      }
      Opcode::StartSpellDifficulty => {
        let (id, time, bonus, name) = decode_args!(args, "SSSp(bs=4;mask=0x77,7,16)");
        Op::StartSpellDifficulty { id, time, bonus, name }
      }
      Opcode::StartSpellDifficultyM1 => {
        let (id, time, bonus, name) = decode_args!(args, "SSSp(bs=4;mask=0x77,7,16)");
        Op::StartSpellDifficultyM1 { id, time, bonus, name }
      }
      Opcode::StartSpellDifficultyM2 => {
        let (id, time, bonus, name) = decode_args!(args, "SSSp(bs=4;mask=0x77,7,16)");
        Op::StartSpellDifficultyM2 { id, time, bonus, name }
      }
      Opcode::SetBossLifeCount => {
        let count = decode_args!(args, "S");
        Op::SetBossLifeCount(param_int(count, 0))
      }
      Opcode::SpawnerReset => {
        let spawner = decode_args!(args, "S");
        Op::SpawnerReset {
          spawner: param_int(spawner, 0),
        }
      }
      Opcode::SpawnerShoot => {
        let spawner = decode_args!(args, "S");
        Op::SpawnerShoot {
          spawner: param_int(spawner, 0),
        }
      }
      Opcode::SpawnerSetSprite => {
        let (spawner, sprite, color) = decode_args!(args, "SSS");
        let (spawner, sprite, color) = (param_int(spawner, 0), param_int(sprite, 1), param_int(color, 2));
        Op::SpawnerSetSprite { spawner, sprite, color }
      }
      Opcode::SpawnerSetOffset => {
        let (spawner, x, y) = decode_args!(args, "Sff");
        let (spawner, x, y) = (param_int(spawner, 0), param_float(x, 1), param_float(y, 2));
        Op::SpawnerSetOffset { spawner, x, y }
      }
      Opcode::SpawnerSetAngle => {
        let (spawner, angle_1, angle_2) = decode_args!(args, "Sff");
        let (spawner, angle_1, angle_2) = (param_int(spawner, 0), param_float(angle_1, 1), param_float(angle_2, 2));
        Op::SpawnerSetAngle {
          spawner,
          angle: [angle_1, angle_2],
        }
      }
      Opcode::SpawnerSetSpeed => {
        let (spawner, speed_1, speed_2) = decode_args!(args, "Sff");
        let (spawner, speed_1, speed_2) = (param_int(spawner, 0), param_float(speed_1, 1), param_float(speed_2, 2));
        Op::SpawnerSetSpeed {
          spawner,
          speed: [speed_1, speed_2],
        }
      }
      Opcode::SpawnerSetCount => {
        let (spawner, count_1, count_2) = decode_args!(args, "SSS");
        let (spawner, count_1, count_2) = (param_int(spawner, 0), param_int(count_1, 1), param_int(count_2, 2));
        Op::SpawnerSetCount {
          spawner,
          count: [count_1, count_2],
        }
      }
      Opcode::SpawnerSetAimMode => {
        let (spawner, mode) = decode_args!(args, "SS");
        let (spawner, mode) = (param_int(spawner, 0), param_int(mode, 1));
        Op::SpawnerSetAimMode { spawner, mode }
      }
      Opcode::SpawnerSetSounds => {
        let (spawner, fire_sound, transform_sound) = decode_args!(args, "SSS");
        let (spawner, fire_sound, transform_sound) = (
          param_int(spawner, 0),
          param_int(fire_sound, 1),
          param_int(transform_sound, 2),
        );
        Op::SpawnerSetSounds {
          spawner,
          fire_sound,
          transform_sound,
        }
      }
      Opcode::SpawnerSetEffects => {
        let (spawner, effect_index, is_async, id, int_1, int_2, float_1, float_2) = decode_args!(args, "SSSSSSff");
        let (spawner, effect_index, is_async, id, int_1, int_2, float_1, float_2) = (
          param_int(spawner, 0),
          param_int(effect_index, 1),
          param_int(is_async, 2),
          param_int(id, 3),
          param_int(int_1, 4),
          param_int(int_2, 5),
          param_float(float_1, 6),
          param_float(float_2, 7),
        );
        Op::SpawnerSetEffects {
          spawner,
          effect_index,
          is_async,
          id,
          int_1,
          int_2,
          float_1,
          float_2,
        }
      }
      Opcode::SpawnerCancel => Op::SpawnerCancel,
      Opcode::SpawnerCopy => {
        let (dest_spawner, src_spawner) = decode_args!(args, "SS");
        let (dest_spawner, src_spawner) = (param_int(dest_spawner, 0), param_int(src_spawner, 1));
        Op::SpawnerCopy {
          dest_spawner,
          src_spawner,
        }
      }
      Opcode::LaserLineCreate
      | Opcode::LaserLineCreateNoBlend
      | Opcode::LaserLineExCreate
      | Opcode::LaserLineExCreateNoBlend
      | Opcode::LaserInfCreate
      | Opcode::LaserInfCreateNoBlend
      | Opcode::LaserInfExCreate
      | Opcode::LaserInfExCreateNoBlend => Op::LaserCreate {
        opcode,
        arguments: args,
      },
      Opcode::BulletCancelRadius => {
        let radius = decode_args!(args, "f");
        Op::BulletCancelRadius(radius)
      }
      Opcode::BulletClearRadius => {
        let radius = decode_args!(args, "f");
        Op::BulletClearRadius(radius)
      }
      Opcode::SpawnerSpeedRank3 => {
        let (spawner, speeds) = spawner_param_decode_args!(args, "S", [[f32; 2]; 3]);
        let spawner = param_int(spawner, 0);
        Op::SpawnerSpeedRank3 { spawner, speeds }
      }
      Opcode::SpawnerSpeedRank5 => {
        let (spawner, speeds) = spawner_param_decode_args!(args, "S", [[f32; 2]; 5]);
        let spawner = param_int(spawner, 0);
        Op::SpawnerSpeedRank5 { spawner, speeds }
      }
      // Opcode::SpawnerSpeedRankLerp => {
      //   let (spawner, speeds) = spawner_param_decode_args!(args, "S", [[f32; 2]; 5]);
      //   let spawner = param_int(spawner, 0);
      //   Op::SpawnerSpeedRankLerp { spawner, speeds }
      // }
      Opcode::SpawnerCountRank3 => {
        let (spawner, counts) = spawner_param_decode_args!(args, "S", [[i32; 2]; 3]);
        let spawner = param_int(spawner, 0);
        Op::SpawnerCountRank3 { spawner, counts }
      }
      Opcode::SpawnerCountRank5 => {
        let (spawner, counts) = spawner_param_decode_args!(args, "S", [[i32; 2]; 5]);
        let spawner = param_int(spawner, 0);
        Op::SpawnerCountRank5 { spawner, counts }
      }
      _ => {
        unimplemented!("opcode {opcode:?} ({opcode_raw}) not implemented");
      }
    };

    Self {
      time: inst.time,
      difficulty: inst.difficulty,
      op,
    }
  }
}
