use std::{io::Cursor, sync::Arc};

use async_std::fs;
use futures::future::{join3, join_all};
use truth::{context::RootEmitter, io::BinReader, Game as TruthGame, StackEclFile};
use wgpu::{naga::FastHashMap, Device, Queue};
use winit::dpi::PhysicalSize;

use crate::{game::anm, utils::game::Game};

use instructions::ecl::Instruction;

pub struct LoadedFile {
  anm_files: Vec<anm::LoadedFile>,
  ecl_files: Vec<LoadedFile>,
  subs: FastHashMap<String, Vec<Instruction>>,
}

impl LoadedFile {
  pub async fn load(
    device: Arc<Device>,
    queue: Arc<Queue>,
    size: PhysicalSize<u32>,
    game: Game,
    file_name: impl Into<String> + Send,
  ) -> LoadedFile {
    let file_name = file_name.into();
    let file_data = fs::read(game.asset_path(&file_name)).await.expect("failed to load anm file");
    let file = StackEclFile::read_from_stream(
      &mut BinReader::from_reader(&RootEmitter::new_stderr(), &file_name, Cursor::new(file_data)),
      match game {
        Game::Mof => TruthGame::Th10,
        Game::Sa => TruthGame::Th11,
      },
    )
    .unwrap();

    let anm_files = join_all(
      file
        .anim_list
        .into_iter()
        .map(|sp| sp.value)
        .map(|anm| async_std::task::spawn(anm::LoadedFile::load(device.clone(), queue.clone(), size, game, anm))),
    );
    let ecl_files = join_all(
      file
        .ecli_list
        .into_iter()
        .map(|sp| sp.value)
        .map(|ecl| LoadedFile::load(device.clone(), queue.clone(), size, game, ecl)),
    );

    let subs = file.subs.into_iter().map(|(k, v)| (k.value, v)).collect::<Vec<_>>();
    let subs = async_std::task::spawn_blocking(|| {
      subs
        .into_iter()
        .map(|(sub, instructions)| {
          let mut offset_instructions: Vec<(_, _)> = Default::default();
          let mut current_offset = 0;
          for (index, inst) in instructions.iter().enumerate() {
            offset_instructions.push((current_offset, index));
            current_offset += 16 + inst.args_blob.len() as u32
          }

          println!("hitting {sub}");
          (
            sub,
            instructions
              .into_iter()
              .zip(offset_instructions.iter())
              .map(|(inst, (offset, _))| Instruction::from_raw(inst, *offset, &offset_instructions))
              .collect::<Vec<_>>(),
          )
        })
        .collect::<FastHashMap<_, _>>()
    });

    let (anm_files, ecl_files, subs) = join3(anm_files, ecl_files, subs).await;

    LoadedFile {
      anm_files,
      ecl_files,
      subs,
    }
  }
}
