use std::{cell::RefCell, ffi::c_void, marker::PhantomData, thread::LocalKey};

#[doc(hidden)]
#[derive(Default)]
pub struct ContextMut<T, Key>(PhantomData<(T, Key)>);

unsafe impl<T, Key> Sync for ContextMut<T, Key> {}

#[allow(unused)]
impl<T, Key> ContextMut<T, Key> {
  pub const fn new() -> ContextMut<T, Key> {
    ContextMut(PhantomData)
  }

  fn thread_local() -> &'static LocalKey<RefCell<Option<*mut c_void>>> {
    std::thread_local! {
      static CTX: RefCell<Option<*mut c_void>> = Default::default();
    }

    &CTX
  }

  pub fn with<U>(&'static self, func: impl FnOnce(&mut T) -> U) -> U {
    Self::thread_local().with_borrow(|borrow| {
      let value = unsafe { &mut *(borrow.expect("attempted to access context while it was vacant") as *mut _) };

      func(value)
    })
  }

  pub fn scoped<'a, U>(&'a self, value: &'a mut T, func: impl 'a + FnOnce() -> U) -> U {
    Self::thread_local().with_borrow_mut(|context| {
      let ptr: *mut T = value;
      assert_eq!(context.replace(ptr as *mut _), None);
    });

    let retval = func();

    Self::thread_local().with_borrow_mut(|context| context.take());

    retval
  }
}

#[doc(hidden)]
#[derive(Default)]
pub struct Context<T, Key>(PhantomData<(T, Key)>);

unsafe impl<T, Key> Sync for Context<T, Key> {}

#[allow(unused)]
impl<T, Key> Context<T, Key> {
  pub const fn new() -> Context<T, Key> {
    Context(PhantomData)
  }

  fn thread_local() -> &'static LocalKey<RefCell<Option<*const c_void>>> {
    std::thread_local! {
      static CTX: RefCell<Option<*const c_void>> = Default::default();
    }

    &CTX
  }

  pub fn with<'a, U>(&'static self, func: impl FnOnce(&'a T) -> U) -> U {
    Self::thread_local().with_borrow(|borrow| {
      let value = unsafe { &mut *(borrow.expect("attempted to access context while it was vacant") as *mut _) };

      func(value)
    })
  }

  pub fn scoped<'a, U>(&'a self, value: &'a T, func: impl FnOnce() -> U) -> U {
    let ptr: *const T = value;
    Self::thread_local().with_borrow_mut(|context| {
      assert_eq!(context.replace(ptr as *const _), None);
    });

    let retval = func();

    Self::thread_local().with_borrow_mut(|context| context.take());

    retval
  }
}
